# Copyright (c) 2002
#   Riverbank Computing Limited <info@riverbankcomputing.co.uk>
#
# This is the build script for PyQt.  It should be run in the top level
# directory of the source distribution and by the Python interpreter for which
# it is being built.  It uses either qmake or tmake to do the hard work of
# generating the platform specific Makefiles.


import sys
import os
import glob
import re
import tempfile
import shutil
import py_compile
import compileall
import string
import time


# Define the globals.
progName = None
qtDir = None
qtIncDir = None
qtVersion = 0
qtLib = None
platPyScriptDir = None
platPyDLLDir = None
platPySiteDir = None
platPyIncDir = None
platPyLib = None
sipModuleDir = None
sipIncDir = None
sipBin = None
sciIncDir = None
sciLibDir = None
sciInstalled = 1
platBinDir = None
modDir = None
platMake = None
platCopy = None
platMkdir = None
platRmdir = None
platQTDIRName = None
pyFullVers = None
makefileGen = None
makeBin = None
buildModules = ["qt"]
proPatches = {}
makefilePatches = {}
tempBuildDir = None
usingTmake = 0
enableOldPython = 0
debugMode = "release"
catCppFiles = 0
qpeTag = None
licType = None
gccFlags = []


def usage(rcode = 2):
    """Display a usage message and exit.

    rcode is the return code passed back to the calling process.
    """
    global progName, platBinDir, platMake, modDir, platQTDIRName, sipIncDir

    print "Usage:"
    print "    %s [-h] [-a version] [-b dir] [-c] [-d dir] [-e dir] [-f gccflag] [-g prog] [-i dir] [-l Qt-library] [-m prog] [-n dir] [-o dir] [-p dir] [-q dir] [-s dir] [-u] [-w]" % (progName)
    print "where:"
    print "    -h             display this help message"
    print "    -a tag         explicitly enable the qtpe module"
    print "    -b dir         where pyuic and pylupdate will be installed [default %s]" % (platBinDir)
    print "    -c             concatenate each module's C++ source files"
    print "    -d dir         where PyQt will be installed [default %s]" % (modDir)
    print "    -e dir         the directory containing the SIP header files [default %s]" % (sipIncDir)
    print "    -f gccflag     additional GCC flag, eg. -fno-exceptions"
    print "    -g prog        the name of the Makefile generator"
    print "    -i dir         the directory containing the Qt header files [default %s%sinclude]" % (platQTDIRName,os.sep)
    print "    -l Qt-library  explicitly specify the type of Qt library, either qt, qt-mt or qte"
    print "    -m prog        the name of the Make program [default %s]" % (platMake)
    print "    -n dir         the directory containing the QScintilla header files [default %s%sinclude]" % (platQTDIRName, os.sep)
    print "    -o dir         the directory containing the QScintilla library [default %s%slib]" % (platQTDIRName, os.sep)
    print "    -p dir         the name of the SIP code generator [default sip]"
    print "    -q dir         the Qt base directory [default %s]" % (platQTDIRName)
    print "    -s dir         the directory containing the SIP module"
    print "    -u             build with debugging symbols"
    print "    -w             enable the use of Python 1.5.x under Windows"

    sys.exit(rcode)


def format(mess,delin = 1):
    """Format a message by inserting line breaks at appropriate places.

    mess is the text of the message.

    delin is non-zero if the text should be deliniated.

    Return the formatted message.
    """
    margin = 78
    curs = 0

    if delin:
        msg = "*" * margin + "\n"
    else:
        msg = ""

    for w in string.split(mess):
        l = len(w)
        if curs != 0 and curs + l > margin:
            msg = msg + "\n"
            curs = 0

        if curs > 0:
            msg = msg + " "
            curs = curs + 1

        msg = msg + w
        curs = curs + l

    msg = msg + "\n"

    if delin:
        msg = msg + ("*" * margin) + "\n"

    return msg


def error(mess):
    """Display an error message and exit.

    mess is the text of the message.
    """
    sys.stderr.write(format("Error: %s" % (mess)))
    sys.exit(1)


def inform(mess,delin = 1):
    """Display an information message.

    mess is the text of the message.

    delin is non-zero if the text should be deliniated.
    """
    sys.stdout.write(format(mess,delin))


def initGlobals():
    """Sets the values of globals that need more than a simple assignment.
    """
    global platMake, platCopy, platPyScriptDir, platPyDLLDir, platPySiteDir
    global platPyIncDir, platPyLib, platQTDIRName, platBinDir, platMkdir
    global platRmdir, modDir, pyFullVers, sipIncDir

    pyFullVers = string.split(sys.version)[0]

    vl = string.split(pyFullVers,".")
    pyVers = vl[0] + "." + vl[1]
    pyVersNr = int(vl[0]) * 10 + int(vl[1])

    if sys.platform == "win32":
        platMake = "nmake"
        platCopy = "copy /y"
        platMkdir = "mkdir"
        platRmdir = "rmdir /s /q"
        platPyScriptDir = sys.prefix + "\\Lib"
        platPyDLLDir = sys.prefix + "\\DLLs"

        if pyVersNr < 22:
            platPySiteDir = platPyScriptDir
        else:
            platPySiteDir = sys.prefix + "\\Lib\\site-packages"

        platPyIncDir = sys.prefix + "\\include"
        platQTDIRName = "%QTDIR%"
        platBinDir = sys.exec_prefix

        platPyLib = sys.prefix + "\\libs\\python%d.lib" % (pyVersNr)
    else:
        vl = string.split(pyFullVers,".")
        pyVers = vl[0] + "." + vl[1]

        platMake = "make"
        platCopy = "cp"
        platMkdir = "mkdir"
        platRmdir = "rm -rf"
        platPyScriptDir = sys.prefix + "/lib/python" + pyVers
        platPyDLLDir = sys.prefix + "/lib/python" + pyVers + "/lib-dynload"

        if pyVersNr < 20:
            platPySiteDir = sys.prefix + "/lib/python" + pyVers
        else:
            platPySiteDir = sys.prefix + "/lib/python" + pyVers + "/site-packages"

        platPyIncDir = sys.prefix + "/include/python" + pyVers
        platQTDIRName = "$QTDIR"
        platBinDir = "/usr/local/bin"

    modDir = platPySiteDir
    sipIncDir = platPyIncDir


def escape(s):
    """Escape all the backslashes (usually Windows directory separators) in a
    string so that they don't confuse regular expression substitutions.  Also
    enclose in double quotes if the string contains spaces.

    s is the string to escape.

    Returns the escaped string.
    """
    s = string.replace(s,"\\","\\\\")

    if s.find(" ") >= 0:
        s = "\"" + s + "\""

    return s


def searchPath(prog,xtradir = None):
    """Search the PATH for a program and return it's absolute pathname.

    prog is the name of the program.
    xtradir is the name of an optional extra directory to search as well.
    """
    if sys.platform == "win32":
        prog = prog + ".exe"

    # Create the list of directories to search.
    try:
        path = os.environ["PATH"]
    except KeyError:
        path = os.defpath

    dlist = string.split(path,os.pathsep)

    if xtradir:
        dlist.insert(0,xtradir)

    # Create the list of files to check.
    flist = []

    for d in dlist:
        flist.append(os.path.abspath(os.path.join(d,prog)))

    # Search for the file.
    for f in flist:
        if os.access(f,os.X_OK):
            return f

    # Nothing found.
    return None


def patchFile(name,pdict,bakfile = None):
    """Apply the patches to a file.

    name is the name of the file.
    pdict is the patch dictionary.
    bakfile is the name of the optional backup file for the unpatched version.
    """
    newname = name + ".new"

    fi = open(name,"r")
    fo = open(newname,"w")

    for line in fi.readlines():
        for (pattern, repl) in pdict.values():
            line = re.sub(pattern,repl,line)

        fo.write(line)
    
    fi.close()
    fo.close()

    if bakfile is None:
        os.remove(name)
    else:
        os.rename(name,bakfile)

    os.rename(newname,name)


def isQtLibrary(dir,lib):
    """See if a Qt library appears in a particular directory.

    dir is the name of the directory.
    lib is the name of the library.

    Returns the basename of the library file that was found or None if nothing
    was found.
    """
    if sys.platform == "win32":
        lpatt = "\\" + lib + "[0-9]*.lib"
    else:
        lpatt = "/lib" + lib + ".*"

    l = glob.glob(dir + lpatt)

    if len(l) == 0:
        return None

    return os.path.basename(l[0])


def copyToFile(name,text):
    """Copy a string to a file.

    name is the name of the file.
    text is the contents to copy to the file.
    """
    f = open(name,"w")
    f.write(text)
    f.close()


def mkTempBuildDir(olddir = None):
    """Create a temporary build directory for a console application called
    qttest, complete with patched Makefile.  The global tempBuildDir is set to
    the name of the directory.  The temporary directory becomes the current
    directory.

    olddir is None if the directory should be created, otherwise it is deleted.

    Returns the name of the previous current directory.
    """
    global tempBuildDir

    if olddir is None:
        tempBuildDir = tempfile.mktemp()

        try:
            os.mkdir(tempBuildDir)
        except:
            error("Unable to create temporary directory.")

        newdir = pushDir(tempBuildDir)

        copyToFile("qttest.pro",
"""TEMPLATE = app
TARGET = qttest
CONFIG = qt @TEST_OPENGL@ console warn_off release @BL_THREAD@
INCLUDEPATH = @BL_INCLUDEPATH@
DEFINES = @BL_DEFINES@
SOURCES = qttest.cpp
LIBS += @TEST_SCILIBDIR@

win32:LIBS += @BL_PYLIB@
""")

        # Disable OpenGL and QScintilla support by default.
        global proPatches
        proPatches["TOPENGL"] = [re.compile("@TEST_OPENGL@",re.M), ""]
        proPatches["TSCILIBDIR"] = [re.compile("@TEST_SCILIBDIR@",re.M), ""]

        # Create a dummy source file to suppress a qmake warning.
        copyToFile("qttest.cpp","")

        buildMakefile("qttest.pro")
    else:
        popDir(olddir)
        newdir = None

        shutil.rmtree(tempBuildDir,1)

    return newdir


def tryModule(mname,incfile,ctor):
    """See if a PyQt module should be built and update the buildModules list
    accordingly.

    mname is the name of the PyQt module.
    incfile is the C++ header file that defines the class being used for the
    test.
    ctor is the constructor of the class being used for the test.
    """
    copyToFile("qttest.cpp",
"""#include <%s>

int main(int argc,char **argv)
{
    new %s;
}
""" % (incfile,ctor))

    if runMake(None,0) == 0:
        global buildModules

        buildModules.append(mname)
        inform("The %s module will be built." % (mname))
    else:
        inform("The %s module will not be built." % (mname))

    runMake("clean")


def checkQtDirAndVersion():
    """Check that Qt can be found and what its version is.
    """
    # Check the Qt directory is known and valid.
    global qtVersion, qtDir, sciInstalled, sciIncDir

    if qtDir is None:
        try:
            qtDir = os.environ["QTDIR"]
        except KeyError:
            error("Please set the name of the Qt base directory, either by using the -q argument or setting the QTDIR environment variable.")
    else:
        os.environ["QTDIR"] = qtDir

    # For lib template project files.
    makefilePatches["QTDIR_LIB"] = [re.compile("(^CC)",re.M), "QTDIR    = " + escape(qtDir) + "\n\\1"]

    # For subdir template project files.
    makefilePatches["QTDIR_SUBDIR"] = [re.compile("(^MAKEFILE)",re.M), "QTDIR    = " + escape(qtDir) + "\n\\1"]

    if not os.access(qtDir,os.F_OK):
        error("The %s Qt base directory does not seem to exist. Use the -q argument or the QTDIR environment variable to set the correct directory." % (qtDir))

    inform("%s is the Qt base directory." % (qtDir))

    # Check that the Qt header files have been installed.
    global qtIncDir

    if qtIncDir:
        qtincdir = qtIncDir
    else:
        qtincdir = qtDir + os.sep + "include"

    qglobal = qtincdir + os.sep + "qglobal.h"

    if not os.access(qglobal,os.F_OK):
        error("qglobal.h could not be found in %s. Use the -i argument to explicitly specify the correct directory." % (qtincdir))

    inform("%s contains qglobal.h." % (qtincdir))

    # Get the Qt version number.
    qtversstr = ""
    f = open(qglobal)
    l = f.readline()

    while l:
        wl = string.split(l)
        if len(wl) == 3 and wl[0] == "#define":
            if wl[1] == "QT_VERSION":
                qtv = wl[2]

                if qtv[0:2] == "0x":
                    qtVersion = string.atoi(qtv,16)
                else:
                    dec = int(qtv)
                    maj = dec / 100
                    min = (dec % 100) / 10
                    bug = (dec % 10)
                    qtVersion = (maj << 16) + (min << 8) + bug

            if wl[1] == "QT_VERSION_STR":
                qtversstr = wl[2][1:-1]

            if qtversstr != "" and qtVersion > 0:
                break

        l = f.readline()

    f.close()

    if qtversstr == "":
        error("The Qt version number could not be determined by parsing %s." % (qglobal))

    # Prevent the Makefiles recreating themselves.  Only seems to affect Qt
    # v3.1.0 and later.
    if qtVersion >= 0x030100:
        makefilePatches["FORCE"] = [re.compile(" FORCE",re.M), ""]

    # Try and work out which Qt edition it is.
    qtedition = ""
    qteditionstr = ""

    if qtVersion >= 0x030000:
        qconfig = qtincdir + os.sep + "qconfig.h"

        if not os.access(qconfig,os.F_OK):
            error("qconfig.h could not be found in %s." % (qtincdir))

        f = open(qconfig)
        l = f.readline()

        while l:
            wl = string.split(l)
            if len(wl) == 3 and wl[0] == "#define" and wl[1] == "QT_PRODUCT_LICENSE":
                    qtedition = wl[2][4:-1]
                    break

            l = f.readline()

        f.close()

        if qtedition == "":
            error("The Qt edition could not be determined by parsing %s." % (qconfig))

        qteditionstr = qtedition + " edition "

    inform("Qt %s %sis being used." % (qtversstr,qteditionstr))

    # Check the licenses are compatible.
    if qtedition:
        global licType

        if (qtedition == "free" and licType != "GPL") or \
           (qtedition != "free" and licType == "GPL"):
            error("This version of PyQt and the %s edition of Qt have incompatible licenses." % (qtedition))

    # The professional edition needs special handling if XML support is needed.
    global proPatches

    if qtedition == "professional":
        proPatches["SCOPE"] = [re.compile("@BL_SCOPE@",re.M), "rbprof"]
    else:
        proPatches["SCOPE"] = [re.compile("@BL_SCOPE@",re.M), ""]

    # Link in the qassistantclient library for Qt v3.1+.
    qaclib = ""

    if qtVersion >= 0x030100:
        if sys.platform == "win32":
            qaclib = "$(QTDIR)\\lib\\qassistantclient.lib"
        else:
            qaclib = "-lqassistantclient"

    proPatches["QACLIB"] = [re.compile("@BL_QASSISTANTCLIENT@",re.M), qaclib]

    # Set the list of directories containing header files.
    incdirs = ". " + escape(platPyIncDir)

    if sipIncDir != platPyIncDir:
        incdirs = incdirs + " " + escape(sipIncDir)

    if qtIncDir:
        incdirs = incdirs + " " + escape(qtIncDir)

    # Find the QScintilla header files.
    if qtVersion >= 0x030000 and sciInstalled:
        if not sciIncDir:
            sciIncDir = qtDir + os.sep + "include"

        if os.access(sciIncDir + os.sep + "qextscintillabase.h",os.F_OK):
            if sciIncDir != qtincdir:
                incdirs = incdirs + " " + escape(sciIncDir)

            inform("%s contains qextscintillabase.h." % (sciIncDir))
        else:
            sciInstalled = 0
            inform("qextscintillabase.h could not be found in %s and so the qtext module will not be built. If QScintilla is installed then use the -n argument to explicitly specify the correct directory." % (sciIncDir))

    proPatches["INCLUDEPATH"] = [re.compile("@BL_INCLUDEPATH@",re.M), incdirs]


def checkQtLibrary():
    """Check which Qt library is to be used.
    """
    global qtDir, qtLib

    qtlibdir = qtDir + os.sep + "lib"

    if qtLib is None:
        mtlib = isQtLibrary(qtlibdir,"qt-mt")
        stlib = isQtLibrary(qtlibdir,"qt")
        emlib = isQtLibrary(qtlibdir,"qte")

        nrlibs = 0
        names = ""

        if stlib:
            nrlibs = nrlibs + 1
            names = names + ", qt"
            qtLib = "qt"
            lib = stlib

        if mtlib:
            nrlibs = nrlibs + 1
            names = names + ", qt-mt"
            qtLib = "qt-mt"
            lib = mtlib

        if emlib:
            nrlibs = nrlibs + 1
            names = names + ", qte"
            qtLib = "qte"
            lib = emlib

        if nrlibs == 0:
            error("No Qt libraries could be found in %s." % (qtlibdir))

        if nrlibs > 1:
            error("These versions of the Qt libraries were found: %s. Use the -l argument to explicitly specify which you want to use." % (names[2:]))

        inform("The %s Qt library was found." % (qtLib))
    else:
        lib = isQtLibrary(qtlibdir,qtLib)

        if not lib:
            error("The %s Qt library could not be found in %s." % (qtLib,qtlibdir))

    if qtLib == "qt-mt":
        proPatches["THREAD"] = [re.compile("@BL_THREAD@",re.M), "thread"]
        inform("Qt thread support is enabled.")
    else:
        proPatches["THREAD"] = [re.compile("@BL_THREAD@",re.M), ""]
        inform("Qt thread support is disabled.")

    # Amazingly, qmake under Windows doesn't know the version of the Qt
    # library, so we have to patch it.
    if sys.platform == "win32":
        global usingTmake

        if usingTmake:
            prolib = "qt.lib"
        else:
            prolib = qtLib + ".lib"

        makefilePatches["QTLIB"] = [re.compile(prolib,re.M), lib]


def installChecks():
    """Carry out basic checks about the installation.
    """
    global proPatches, makefilePatches, pyFullVers

    # Check we are in the top level directory.
    global progName

    if not os.access(progName,os.F_OK):
        error("This program must be run from the top level directory of the package, ie. the directory containing this program.")

    global licType

    try:
        import license
        licType = license.LicenseType
        licname = license.LicenseName
    except:
        licType = "GPL"
        licname = "GNU General Public License"

    inform("This is the %s version of PyQt and is licensed under the %s." % (licType,licname),0)

    print
    print "Type 'L' to view the license."
    print "Type 'yes' to accept the terms of the license."
    print "Type 'no' to decline the terms of the license."
    print

    while 1:
        try:
            resp = raw_input("Do you accept the terms of the license? ")
        except:
            resp = ""

        resp = resp.strip().lower()

        if resp == "yes":
            break

        if resp == "no":
            sys.exit(0)

        if resp == "l":
            os.system("more LICENSE")

    inform("Building the %s version of PyQt 3.5 for Python %s on %s." % (licType,pyFullVers,sys.platform))

    if licType != "GPL":
        if os.access("sip/" + license.LicenseFile,os.F_OK):
            inform("Found the license file %s.\n" % (license.LicenseFile))
        else:
            error("Please copy the license file %s to the sip directory.\n" % (license.LicenseFile))

    # We can't be sure that we can run programs under Windows (because it might
    # be 95 or 98) with Python 1.5.x and get the return codes.  So we force the
    # user to explicitly say that's what they want to do.
    if sys.platform == "win32" and int(pyFullVers[0]) == 1 and int(pyFullVers[2]) < 6:
        global enableOldPython

        if not enableOldPython:
            error("You are using Python %s under Windows. If you are running Windows95 or Windows98 then there may be problems trying to configure PyQt. It is recomended that you upgrade to Python 1.6 or later. Use the -w argument to suppress this check.\n" % (pyFullVers))

    # Check the installation directory is valid.
    global modDir

    if not os.access(modDir,os.F_OK):
        error("The %s PyQt destination directory does not seem to exist. Use the -d argument to set the correct directory." % (modDir))

    proPatches["DESTDIR"] = [re.compile("@BL_DESTDIR@",re.M), escape(modDir)]

    inform("%s is the PyQt installation directory." % (modDir))

    # Check that the Python header files have been installed.
    global platPyIncDir

    if not os.access(platPyIncDir + "/Python.h",os.F_OK):
        error("Python.h could not be found in %s. Make sure Python is fully installed, including any development packages." % (platPyIncDir))

    inform("%s contains Python.h." % (platPyIncDir))

    # Check that the Python library can be found.
    global platPyLib

    if platPyLib:
        if not os.access(platPyLib,os.F_OK):
            error("%s could not be found. Make sure Python is fully installed, including any development packages." % (platPyLib))

        proPatches["PYLIB"] = [re.compile("@BL_PYLIB@",re.M), escape(platPyLib)]

        inform("%s is installed." % (platPyLib))

    # Look for the SIP binary.
    global sipBin

    if sipBin is None:
        global platBinDir

        sipBin = searchPath("sip",platBinDir)
    elif not os.access(sipBin,os.X_OK):
        sipBin = None

    if sipBin is None:
        error("The SIP code generator could not be found. Use the -p argument to set the correct program.")

    inform("%s will be used as the SIP code generator." % (sipBin))

    # Check that the SIP header files have been installed.
    global sipIncDir

    if not os.access(sipIncDir + "/sipQt.h",os.F_OK):
        error("sipQt.h could not be found in %s. Use the -e argument to set the correct directory." % (sipIncDir))

    inform("%s contains sipQt.h." % (sipIncDir))

    global sipModuleDir

    if sipModuleDir is None:
		global platPyScriptDir, platPyDLLDir, platPySiteDir

		dirlist = [platPyDLLDir, platPyScriptDir, platPySiteDir]
    else:
        dirlist = [sipModuleDir]
        sipModuleDir = None

    for d in dirlist:
        if glob.glob(d + "/libsip*"):
            sipModuleDir = d
            break

    if sipModuleDir is None:
        error("The SIP module/library could not be found. Use the -s argument to set the correct directory.")

    proPatches["LIBS"] = [re.compile("@BL_SIPMODDIR@",re.M), escape(sipModuleDir)]

    inform("%s contains the SIP module." % (sipModuleDir))

    # Check the Qt installation.
    checkQtDirAndVersion()

    # Look for the Makefile generator.
    checkMakefileGenerator()

    if usingTmake and sys.platform == "win32":
        defines = "QT_DLL"
    else:
        defines = ""

    proPatches["DEFINES"] = [re.compile("@BL_DEFINES@",re.M), defines]

    checkQtLibrary()

    # Look for Make.
    global makeBin, platMake

    if makeBin is None:
        makeBin = searchPath(platMake)
    elif not os.access(makeBin,os.X_OK):
        makeBin = None

    if makeBin is None:
        error("A make program could not be found. Use the -m argument to set the correct program.")

    # For lib template project files.
    makefilePatches["MAKE_LIB"] = [re.compile("(^CC)",re.M), "MAKE     = " + escape(makeBin) + "\n\\1"]

    # For subdir template project files.
    makefilePatches["MAKE_SUBDIR"] = [re.compile("(^MAKEFILE)",re.M), "MAKE     = " + escape(makeBin) + "\n\\1"]

    inform("%s will be used as the make program." % (makeBin))


def checkMakefileGenerator():
    """Check that the Makefile generator can be found.
    """
    global makefileGen, usingTmake

    if makefileGen is None:
        global qtVersion

        if qtVersion >= 0x030000:
            mfg = "qmake"
        else:
            mfg = "tmake"

        xtradir = qtDir + "/bin"

        makefileGen = searchPath(mfg,xtradir)
    elif not os.access(makefileGen,os.X_OK):
        mfg = makefileGen
        makefileGen = None

    if makefileGen is None:
        error("The Makefile generator %s could not be found. Use the -g argument to set the correct program." % (mfg))

    if os.path.basename(makefileGen) in ("tmake", "tmake.exe"):
        usingTmake = 1

        # Make sure that references to tmake use the correct absolute pathname.
        makefilePatches["TMAKE1"] = [re.compile("(^TMAKE[ \t]*=).*$",re.M), "\\1 " + escape(makefileGen)]
        makefilePatches["TMAKE2"] = [re.compile("^\ttmake(.*)$",re.M), "\t" + escape(makefileGen) + "\\1"]

        # Add the install target for Makefiles generated from the subdirs
        # template.
        makefilePatches["TMAKE3"] = [re.compile("^(clean release debug):$",re.M), "\\1 install:"]

        # Remove the invocation of the tmake_all target.
        makefilePatches["TMAKE4"] = [re.compile(" tmake_all",re.M), ""]
    else:
        # Make sure that references to qmake use the correct absolute pathname.
        makefilePatches["QMAKE1"] = [re.compile("(^QMAKE[ \t]*=).*$",re.M), "\\1 " + escape(makefileGen)]

        # Remove the invocation of the qmake_all target.
        makefilePatches["QMAKE2"] = [re.compile(" qmake_all",re.M), ""]

    inform("%s will be used to generate Makefiles." % (makefileGen))

    # If it is tmake then make sure TMAKEPATH is set.
    if usingTmake:
        try:
            os.environ["TMAKEPATH"]
        except:
            error("tmake requires that the TMAKEPATH environment variable is properly set.")


def tuneCompiler():
    """Tune the compiler flags.
    """
    global gccFlags

    if len(gccFlags) > 0:
        subst = "\\1"

        for f in gccFlags:
            subst = subst + " -f" + f

        subst = subst + "\n"

        global makefilePatches

        makefilePatches["CXXFLAGS"] = [re.compile("(^CXXFLAGS.*)$",re.M), subst]


def compileChecks():
    """Tune the compiler flags, and see which PyQt modules to build.
    """
    tuneCompiler()

    # Check which PyQt modules will build.
    if qtVersion >= 0x020000:
        inform("Checking which additional PyQt modules to build.");

        tryModule("qtcanvas","qcanvas.h","QCanvas()")
        tryModule("qtnetwork","qsocket.h","QSocket()")
        tryModule("qttable","qtable.h","QTable()")
        tryModule("qtxml","qdom.h","QDomImplementation()")

        if qtVersion >= 0x030000:
            tryModule("qtsql","qsql.h","QSql()")

        # We need a different Makefile for the qtgl module.
        global proPatches
        proPatches["TOPENGL"] = [re.compile("@TEST_OPENGL@",re.M), "opengl"]
        buildMakefile("qttest.pro")

        tryModule("qtgl","qgl.h","QGLWidget()")

        # Put things back.
        proPatches["TOPENGL"] = [re.compile("@TEST_OPENGL@",re.M), ""]
        buildMakefile("qttest.pro")

        # Check the location of the QScintilla library.
        global sciInstalled

        if qtVersion >= 0x030000 and sciInstalled:
            global sciLibDir

            if not sciLibDir:
                global qtDir
                sciLibDir = qtDir + os.sep + "lib"

            if sys.platform == "win32":
                lpatt = "\\qscintilla.lib"
            else:
                lpatt = "/libqscintilla.*"

            if len(glob.glob(sciLibDir + lpatt)) == 0:
                inform("The QScintilla library could not be found in %s and so the qtext module will not be built. If QScintilla is installed then use the -o argument to explicitly specify the correct directory." % (sciLibDir))
            else:
                proPatches["SCILIBDIR"] = [re.compile("@BL_SCILIBDIR@",re.M), escape(sciLibDir)]

                inform("%s contains the QScintilla library." % (sciLibDir))

                # We need a different Makefile for the qtext module.
                if sys.platform == "win32":
                    scilib = escape(sciLibDir + "\\qscintilla.lib")
                else:
                    scilib = escape("-L" + sciLibDir) + " -lqscintilla"

                proPatches["TSCILIBDIR"] = [re.compile("@TEST_SCILIBDIR@",re.M), scilib]
                buildMakefile("qttest.pro")

                tryModule("qtext","qextscintillabase.h","QextScintillaBase()")

                # Put things back.
                proPatches["TSCILIBDIR"] = [re.compile("@TEST_SCILIBDIR@",re.M), ""]
                buildMakefile("qttest.pro")



def generateFeatures(featfile):
    """Generate the header file describing the Qt features that are enabled if
    it doesn't already exist.  (If it already exists then we are probably cross
    compiling and generated the file through other means.)

    featfile is the name of the features file.
    """
    if os.access(featfile,os.F_OK):
        inform("Using existing features file.")
        return

    inform("Generating the features file.")

    # The features that a given Qt configuration may or may not support.
    flist = ["ACTION", "CLIPBOARD", "CODECS", "COLORDIALOG", "DATASTREAM",
             "DIAL", "DNS", "DOM", "DRAGANDDROP", "ICONVIEW", "IMAGE_TEXT",
             "INPUTDIALOG", "FILEDIALOG", "FONTDATABASE", "FONTDIALOG",
             "MESSAGEBOX", "MIMECLIPBOARD", "NETWORKPROTOCOL", "PICTURE",
             "PRINTDIALOG", "PRINTER", "PROGRESSDIALOG", "PROPERTIES",
             "SEMIMODAL", "SIZEGRIP", "SOUND", "SPLITTER", "STYLE_CDE",
             "STYLE_INTERLACE", "STYLE_MOTIF", "STYLE_MOTIFPLUS",
             "STYLE_PLATINUM", "STYLE_SGI", "STYLE_WINDOWS", "TABDIALOG",
             "TABLE", "TABLEVIEW", "TRANSFORMATIONS", "TRANSLATION", "WIZARD",
             "WORKSPACE"]

    # Generate the program which will generate the features file.
    f = open("qttest.cpp","w")

    # Escape the backslashes so that the same can be embedded in a C++ string.
    ffstr = string.replace(featfile,"\\","\\\\")

    f.write(
"""#include <stdio.h>
#include <Python.h>
#include <qglobal.h>
#include <qapplication.h>

int main(int argc,char **argv)
{
    FILE *fp;
    QApplication app(argc,argv,0);

    if ((fp = fopen("%s","w")) == NULL)
    {
        printf("Unable to create '%s'\\n");
        return 1;
    }

#if !defined(QT_THREAD_SUPPORT) || !defined(WITH_THREAD)
    fprintf(fp,"-x Qt_THREAD_SUPPORT\\n");
#endif
""" % (ffstr, ffstr))

    for feat in flist:
        f.write(
"""
#if defined(QT_NO_%s)
    fprintf(fp,"-x Qt_%s\\n");
#endif
""" % (feat,feat))

    f.write(
"""
    fclose(fp);

    return 0;
}
""")

    f.close()

    runMake()
    runProgram(os.getcwd() + os.sep + "qttest")
    runMake("clean")

    inform("Generated the features file.")


def generateSource(mname,plattag,qttag,xtrtag = None):
    """Generate the C++ source code for a particular PyQt module.

    mname is the name of the module.
    plattag is the SIP tag for the platform.
    qttag is the SIP tag for the Qt version.
    xtrtag is an optional extra SIP tag.
    """
    inform("Generating the C++ source for the %s module." % (mname))

    try:
        shutil.rmtree(mname)
    except:
        pass

    try:
        os.mkdir(mname)
    except:
        error("Unable to create the %s directory." % (mname))

    pro = mname + ".pro"

    argv = ["-t", plattag,
            "-t", qttag,
            "-z", "features",
            "-I", "sip",
            "-m", mname + "/" + pro,
            "-c", mname,
            "sip/" + mname + "mod.sip"]

    if xtrtag:
        argv.insert(0,xtrtag)
        argv.insert(0,"-t")

    global sipBin
    runProgram(sipBin,argv)

    # Python names binary modules differently on different platforms.
    global proPatches, makefilePatches

    target = mname + "c"

    if sys.platform == "win32":
        target = "lib" + target

        # We want to change the module extension on Windows.
        makefilePatches["TARGET"] = [re.compile(target + "\\.dll",re.M), target + ".pyd"]
    else:
        target = target + "module"

    # Generate the Makefile.
    inform("Generating the Makefile for the %s module." % (mname))

    olddir = pushDir(mname)

    proPatches["TARGET"] = [re.compile("@BL_TARGET@",re.M), target]

    global debugMode
    proPatches["DEBUG"] = [re.compile("@BL_DEBUG@",re.M), debugMode]

    global catCppFiles
    buildMakefile(pro,catCppFiles)

    fixInstallTarget(mname)
    del proPatches["TARGET"]

    if sys.platform == "win32":
        del makefilePatches["TARGET"]

    # Compile the Python part of the module.
    pyname = mname + ".py"

    inform("Compiling %s." % (pyname))
    py_compile.compile(pyname)

    popDir(olddir)


def pushDir(newdir):
    """Change to a new directory and return the old one.

    newdir is the new directory.
    """
    olddir = os.getcwd()
    os.chdir(newdir)

    return olddir


def popDir(olddir):
    """Change back to a previous directory.

    olddir is the old directory.
    """
    os.chdir(olddir)


def runProgram(prog,argv = [],fatal = 1):
    """Execute a program.

    prog is the name of the program.
    argv is the optional list of program arguments.
    fatal is non-zero if an error is considered fatal.

    Returns the error code.
    """
    # Use the basename to avoid problems with pathnames containing spaces.
    argv.insert(0,os.path.basename(prog))

    try:
        rc = os.spawnv(os.P_WAIT,prog,argv)
    except AttributeError:
        argv[0] = prog
        rc = os.system(string.join(argv))
    except:
        raise

    if rc != 0 and fatal:
        error("%s failed with an exit code of %d." % (prog,rc))

    return rc


def runMake(target = None,fatal = 1):
    """Run make.

    target is the optional make target.
    fatal is non-zero if an error is considered fatal.

    Returns the error code.
    """
    global makeBin

    if target is None:
        argv = []
    else:
        argv = [target]

    # As the tests are quite quick, make sure that we don't get timestamp
    # problems.
    time.sleep(1)

    return runProgram(makeBin,argv,fatal)


def buildMakefile(pro,catcpp = 0):
    """Run qmake or tmake to create a Makefile from a project file.

    pro is the name of the project file.
    catcpp is non-zero if the project file is for a module whose C++ files are
    to be concatenated.
    """
    global makefileGen, proPatches, makefilePatches

    # Preserve the original project file.
    probak = pro + ".bak"
    patchFile(pro,proPatches,probak)

    if catcpp:
        catFiles(pro)

    # Make sure that the Makefile has a later timestamp than the project file.
    time.sleep(1)

    runProgram(makefileGen,['-o', 'Makefile', pro])
    os.remove(pro)
    os.rename(probak,pro)

    patchFile("Makefile",makefilePatches)


def catFiles(pro):
    """Concatenate a modules C++ source files.

    pro is the name of the project file.
    """
    # Extract the module name from the project file name.
    mname = os.path.splitext(pro)[0]

    inform("Concatenating the C++ files for the %s module." % (mname))

    f = open(pro,"r")
    buf = f.read()
    f.close()

    pat = re.compile("^SOURCES =.*.cpp\n",re.M | re.S)
    match = pat.search(buf)
    srclist = buf[match.start(0):match.end(0)]
    srclist = srclist[13:-1]
    srclist = string.replace(srclist,"\\\n\t","")
    srclist = string.split(srclist," ")

    # Concatenate the files.
    d = open(mname + "huge.cpp","w")

    for cppfile in srclist:
        f = open(cppfile,"r")
        d.write(f.read())
        f.close()

    d.close()

    # Replace the C++ file names in the project file.
    buf = re.sub(pat,"SOURCES = " + mname + "huge.cpp\n",buf)

    f = open(pro + ".new","w")
    f.write(buf)
    f.close()

    os.remove(pro)
    os.rename(pro + ".new",pro)


def fixInstallTarget(mname = None):
    """Fix the install target for a Makefile.

    mname is the name of the PyQt module, or None if the target isn't for a
    module.
    """
    f = open("Makefile","a")
    f.write("\ninstall:\n")

    if mname:
        global modDir, platCopy

        f.write("\t-%s %s.py %s\n" % (platCopy,mname,modDir))
        f.write("\t-%s %s.pyc %s\n" % (platCopy,mname,modDir))

        if sys.platform == "win32":
            f.write("\t-del %s\\lib%sc.lib\n" % (modDir,mname))
            f.write("\t-del %s\\lib%sc.exp\n" % (modDir,mname))

    f.close()


def createPyWrapper(pydir,dname,wfile):
    """Create an executable wrapper for a Python script.

    pydir is the name of the directory where the Python script will eventually
    be installed.
    dname is the name of the directory containing the Python script and where
    the wrapper will be created.
    wfile is the basename of the wrapper.

    Returns the platform specific name of the wrapper.
    """
    if sys.platform == "win32":
        wname = dname + "\\" + wfile + ".bat"
        wrapper = \
"""@%s\\pythonw %s\\%s.py %%1 %%2 %%3 %%4 %%5 %%6 %%7 %%8 %%9
""" % (platBinDir,pydir,wfile)
    else:
        wname = dname + "/" + wfile
        wrapper = \
"""#!/bin/sh

exec python %s/%s.py $*
""" % (pydir,wfile)

    copyToFile(wname,wrapper)
    os.chmod(wname,0755)

    return wname


def main(argv):
    """The main function of the script.

    argv is the list of command line arguments.
    """
    import getopt

    # Parse the command line.
    global progName
    progName = os.path.basename(argv[0])

    initGlobals()

    try:
        optlist, args = getopt.getopt(argv[1:],"ha:b:cd:e:f:g:i:l:m:n:o:p:q:s:uw")
    except getopt.GetoptError:
        usage()

    global makeBin, platBinDir, qtLib, qpeTag

    explicitMake = 0

    for opt, arg in optlist:
        if opt == "-h":
            usage(0)
        elif opt == "-a":
            qpeTag = arg
        elif opt == "-b":
            platBinDir = arg
        elif opt == "-c":
            global catCppFiles
            catCppFiles = 1
        elif opt == "-d":
            global modDir
            modDir = arg
        elif opt == "-e":
            global sipIncDir
            sipIncDir = arg
        elif opt == "-f":
            global gccFlags
            gccFlags.append(arg)
        elif opt == "-g":
            global makefileGen
            makefileGen = arg
        elif opt == "-i":
            global qtIncDir
            qtIncDir = arg
        elif opt == "-l":
            if arg in ("qt", "qt-mt", "qte"):
                qtLib = arg
            else:
                usage()
        elif opt == "-m":
            makeBin = arg
            explicitMake = 1
        elif opt == "-n":
            global sciIncDir
            sciIncDir = arg
        elif opt == "-o":
            global sciLibDir
            sciLibDir = arg
        elif opt == "-p":
            global sipBin
            sipBin = arg
        elif opt == "-q":
            global qtDir
            qtDir = arg
        elif opt == "-s":
            global sipModuleDir
            sipModuleDir = arg
        elif opt == "-u":
            global debugMode
            debugMode = "debug"
        elif opt == "-w":
            global enableOldPython
            enableOldPython = 1

    installChecks()
    maindir = mkTempBuildDir()
    compileChecks()

    # Work out the platform and Qt version tags to pass to SIP to generate the
    # code we need.
    if qtLib == "qte":
        plattag = "WS_QWS"
    elif sys.platform == "win32":
        plattag = "WS_WIN"
    else:
        plattag = "WS_X11"

    qttags = {
        0x020000: "Qt_1_43",
        0x020100: "Qt_2_00",
        0x020200: "Qt_2_1_0",
        0x020300: "Qt_2_2_0",
        0x020301: "Qt_2_3_0",
        0x030000: "Qt_2_3_1",
        0x030001: "Qt_3_0_0",
        0x030002: "Qt_3_0_1",
        0x030004: "Qt_3_0_2",
        0x030005: "Qt_3_0_4",
        0x030006: "Qt_3_0_5",
        0x030100: "Qt_3_0_6",
        0x030101: "Qt_3_1_0",
        0x040000: "Qt_3_1_1"
    }

    global qtVersion
    qttag = None

    vl = qttags.keys()
    vl.sort()

    for v in vl:
        if qtVersion < v:
            qttag = qttags[v]
            break

    if qttag is None:
        error("Invalid Qt version: 0x%06x." % (qtVersion))

    # Generate the features file.
    generateFeatures(maindir + os.sep + "features")

    # We don't need the temporary build directory anymore.
    mkTempBuildDir(maindir)

    global buildModules
    subdirs = ""
    for mname in buildModules:
        generateSource(mname,plattag,qttag)
        subdirs = subdirs + " " + mname

    # We handle the qtpe module explicitly rather than auto-detect.  This is
    # because it does things a bit differently and I'm too lazy to deal with it
    # properly at the moment.
    if qpeTag:
        generateSource("qtpe",plattag,qttag,qpeTag)
        subdirs = subdirs + " qtpe"

    # See which version of pyuic to build.
    global proPatches
    proPatches["BINDIR"] = [re.compile("@BL_BINDIR@",re.M), escape(platBinDir)]

    if qtVersion >= 0x030000:
        inform("Creating Makefile for pyuic3.")
        subdirs = subdirs + " pyuic3"
        olddir = pushDir("pyuic3")
    elif qtVersion >= 0x020000:
        inform("Creating Makefile for pyuic2.")
        subdirs = subdirs + " pyuic2"
        olddir = pushDir("pyuic2")

    buildMakefile("pyuic.pro")
    fixInstallTarget()
    popDir(olddir)

    # Build pylupdate if Qt v3.0 or later.
    if qtVersion >= 0x030000:
        inform("Creating Makefile for pylupdate3.")
        subdirs = subdirs + " pylupdate3"
        olddir = pushDir("pylupdate3")

        buildMakefile("pylupdate.pro")
        fixInstallTarget()
        popDir(olddir)

    # Generate the top-level Makefile.
    inform("Creating top level Makefile.")
    copyToFile("PyQt.pro","TEMPLATE = subdirs\nSUBDIRS =" + subdirs + "\n")
    buildMakefile("PyQt.pro")

    # Tell the user what to do next.
    if explicitMake:
        mk = makeBin
    else:
        (mk,ignore) = os.path.splitext(os.path.basename(makeBin))

    msg = "The build of the PyQt source code for your system is now complete. To compile and install PyQt run \"%s\" and \"%s install\" with appropriate user privileges." % (mk,mk)

    if qtVersion >= 0x030000:
        msg = msg + " eric the IDE written using PyQt is now available separately from http://www.die-offenbachs.de/detlev/eric3.html."

    inform(msg)


if __name__ == "__main__":
    try:
        main(sys.argv)
    except SystemExit:
        raise
    except:
        print \
"""An internal error occured.  Please report all the output from the program,
including the following traceback, to support@riverbankcomputing.co.uk.
"""
        raise
